# DEPRECATED
# This module is an alias to the server module and should not be used.
# It is temporarily left in place so that existing references do not break.
from .server import *

print("""
The 'server_api_examples' module is deprecated! Use the 'server' module
directly going forward.
""")
